/*				MINLOG.H
*
*	This is the header for the minimal logic theorem prover
*	minlog.c.  It is included in all constituent .c files.
*/

#include <stdio.h>
#include <ctype.h>


#define MAXSUBS 64
#define IBMAX 500
#define BRANCH_MAX 500
#define HMAX 10000
#define HOF 50

#define Bot 1

#define is_atom(x) (( 'a' <= x && x <= 'z' ) || x == 'F' )
#define atomic(x) (0 < x && x < max_atoms)

typedef enum { false, true } boolean;

typedef enum { norule, inspect, BotE, iI, aE, MP, TNe, or_if, aI, vE, vI, iE, 
		iE2n, has, ASS, REIT, n_rules } rule;

typedef enum { try, member, absurd, bot, ifI, andE, MPP, TN, redundant,
		share, andI, orE, gotT, gotF, B2F, orIT, orIF, inbr, CP1F, 
		ifE1, CP2F, ifE2, nogood, max_event } event;

#define halfset unsigned

typedef struct S_E_T_
{	halfset lowbits;
	halfset highbits;
} SET, *set;

#define WITHALL(x) for (x = 0; x < max_subs; x++) if (in( x, Gamma ))

#define WITH(x) for (x=max_atoms; x<max_subs; x++) if (in( x, Gamma ))

#define P_WITHALL(px,x) for (px=0; x=premises[px]; px++)

#define P_WITH(px,x) P_WITHALL(px,x) if ( !atomic(x) )

#define DUMP_SEQUENT(s) { printf("%-6s  ",s); \
	PDeparse(Gamma); printf("  |-  "); pdeparse(Con); puts(""); }

/*
*	Now come two types for representation of formulas.
*	The first is for I/O and for the prover proper, while
*	the second is for the tester (as in the case of MaGIC).
*/

typedef struct W_F_F
{	char	sym;
	int	left, right;
} WFF;

typedef struct TW_F_F
{	int	*mtx,		/* Start of the relevant matrix 	*/
		*lv, *rv,	/* To values of subformulas		*/
		val;		/* Currently assigned value		*/
	struct TW_F_F
		*nxt;		/* Next wff to be assigned a value	*/
} TWFF;


typedef struct P_R_O_O_F
{	rule	rl;		/* Last rule used in proof		*/
	int	form,		/* Main formula to which rule applied	*/
		conc;		/* Desired conclusion at the time	*/
	struct P_R_O_O_F
		*lsprf,
		*rsprf;		/* Pointers to left and right subproofs	*/
} PROOF, *proof;

/*
*	There is a simple binary tree data type for the databases
*	of previously proved and previously disproved sequents.
*/

typedef struct T_R_E_E
{	set	prem;			/* Premises of a sequent	*/	
	int	conc;			/* Conclusion of same		*/
	struct T_R_E_E
		*leftsub, *rightsub;	/* Binary tree pointers		*/
} TREE;


/*
*	It is now necessary to define the global variables.
*	These are shared by functions in different files.
*/

#ifdef MAINFILE
#define GLOBAL
#else
#define GLOBAL extern
#endif


GLOBAL WFF wff[MAXSUBS];	/* Subformulas of the sequent		*/

GLOBAL TWFF subf[MAXSUBS];	/* Ditto as required by the tester	*/

GLOBAL TREE *valids, *invalids;	/* Roots of database trees		*/

GLOBAL halfset
	v_used[MAXSUBS],	/* Atoms occurring in subformulas	*/
	used,			/* Atomic formulas used in sequent	*/
	p_used,			/* Subset of used occurring in premises	*/
	c_used,			/* Ditto the conclusion			*/
	p_mask;			/* Zero for atoms, One for all others	*/

GLOBAL int
	max_atoms,		/* Actual number of atoms occurring	*/
	max_subs,		/* Actual number of subformulas		*/
	seq_number,		/* Sequent counter			*/	
	zero,			/* Zero!!				*/
	heap_ptr,		/* Number of subproofs active		*/
	valid_count,		/* Number of proofs found so far	*/
	invalid_count,		/* Number of disproofs found		*/
	ibmax,			/* Number of sequents in the branch	*/
	verbosity,		/* Amount of runtime statistics printed	*/
	event_counter;		/* The number of events logged		*/

GLOBAL int
	con[2][2],		/* The B2 conjunction matrix		*/
	dis[2][2],		/* The B2 disjunction matrix		*/
	aro[2][2],		/* The B2 implication matrix		*/
	c_branch[BRANCH_MAX];	/*  ->|- applications			*/

GLOBAL SET p_branch[BRANCH_MAX];	/*  Currently active		*/

GLOBAL event
	profile[max_event],	/* Log of various events		*/
	prof_sum[max_event];	/* Cumulative version of profile	*/

GLOBAL char
	*event_string[max_event],	/* Messages for events		*/
	*rule_name[n_rules+1];		/* As it says, rule names	*/

GLOBAL PROOF Heap[HMAX];	/* Nodes of proof trees			*/

GLOBAL unsigned
	hall_of_fame[4][HOF],	/* Hardest cases so far			*/
	fives[MAXSUBS];		/* Semantics for 5-variable case	*/

GLOBAL boolean
	good_sequent,			/* What it says			*/
	intuitionist,			/* Bot entails everything	*/
        graphs, 			/* Make tex pictures of profile	*/
	printem;			/* Print proofs as found	*/

/*
*	Now to declare all the functions with parameter types
*/

/*
*	In file initial.c:
*/
void prog_initial( int argc, char *argv[] );
void initialise();
void subtract( TREE *record );
boolean got_sequent( set Gamma, int *conclusion );
boolean insequent( set Gamma, int *conclusion );
void set_up_wff( set Gamma, int *conclusion );
void exchange( int lower, int higher, set Gamma, int *conclusion );
void set_up_subf();
void set_up_v_used();
void dump_wff( set, int );
void Abort( char * );

/*
*	In file parser.c:
*/
char *dyadic( char c );
int parse( char fla[] );
int s_parse( char fla[] );
int Match( char *start );
int Finish( int subf[], char conn[], int tot );
int Loc( char mn, int lft, int rgt );

/*
*	In file prover.c:
*/
proof proved( set Gamma, int conclusion, set Gamma2, set Extras );

/*
*	In file trees.c
*/
boolean got( TREE *root, set Gamma, int conclusion );
void add( boolean good, set Gamma, int conclusion );
void r_add( TREE *root, set Gamma, int conclusion );
TREE *newguy( set Gamma, int conclusion );
boolean inbranch( set Gamma, int conclusion );
void putinbranch( set Gamma, int conclusion );
void outbranch( set Gamma, int conclusion );

/*
*	In file tester.c:
*/
void set_used( set Gamma, int conclusion );
halfset r_set_used( int formula );
boolean H3_valid( int premises[MAXSUBS], int conclusion, set Sigma );
boolean B2_valid( int premises[MAXSUBS], int conclusion, set Sigma );
boolean B2_valid_test( int premises[MAXSUBS], int conclusion, set Gamma );

/*
*	In file events.c
*/

void events_init();
void report( event e, int nn );
void dump_events( event p[max_event] );
void Deparse( set );
void deparse( int );
void sub_deparse( boolean, int );
void PDeparse( set );
void pdeparse( int );
void psub_deparse( int, boolean );
void proof_out( proof P, set Gamma );
void print_proof( proof P, int indent );
void print_node( int fmla, rule rul, int indent );
void blank_line( int indent );
