/*				ML_PARSER.C
*
*	This parser is loosely based on that used by MaGIC and
*	like programs.  The differences are that the stock of
*	connectives is fixed, that negation is paraphrased away
*	since this is minimal logic, and that the value returned
*	is an integer.  Formula 0 is the dummy.
*/


#include "minlog.h"


#define _P_OK (c && c!=')')

#define Error(s,ws) { *_P_ERR=s; _P_ERR[1]=ws; return 0; }

static char
	dyads[6] = { '*', '&', '+', 'V', '>', '\0' },
	*_P_ERR[2];



char *dyadic( char c )
{
	int i;
	for ( i = 0; dyads[i]; i++ )
	if ( dyads[i] == c ) return dyads+i;
	return 0;
}



int parse( char fla[] )
{
	int i = 0, j = 0;

	while ( fla[i] ) if ( fla[i++] != ' ' )
	fla[j++] = fla[i-1];  fla[j] = '\0';
	if ( !*fla ) return 0;

	i = s_parse( fla );

	if ( i ) return i;

	printf("\n ERROR: "); puts(*_P_ERR);
	for ( i = 0; i < _P_ERR[1]-fla; i++ ) printf(" ");
	puts("!");  puts(fla);
	return 0;
}




int s_parse( char fla[] )
{
	int sbf[20],		/* Subformulas in present parentheses	*/
	pptr = 0,		/* Offset of current byte		*/
	oldptr,			/* Pptr before sequence of '~'s 	*/
	ns = 0,			/* Number of subformulas in sbf		*/
	i,j;
	char dcn[19],		/* Dyadic connectives, current bracket	*/
	c;			/* Value of current byte		*/

	c = *fla;
	while (_P_OK) {
		oldptr = pptr;
		while ( (c=fla[pptr++]) == '~' ); 
		i = pptr-1;		if ( is_atom(c) )
		sbf[ns] = Loc( c, 0, 0 );
		else if (c == '(' || c == '.') {
			if ( !(sbf[ns] = s_parse( fla+pptr)) ) return 0;
			j = Match( fla+pptr-1 );
			if ( !j ) return 0;
			pptr += j-1;
		}
		else Error("This can't begin a subformula", fla+pptr-1)
		if ( fla[oldptr] == '~' )
			while ( i-- > oldptr )
				sbf[ns] = Loc('>', sbf[ns], 1);
		c = fla[pptr];
		if (_P_OK) {
			if (c == '-' && fla[pptr+1] == '>') c = fla[++pptr];
			if ( dyadic(c) ) dcn[ns++] = c; else {
				printf("Bad character %c\n", c);
				Error("Dyadic connective expected here", fla+pptr)
			}
			c = fla[++pptr];
			if (!_P_OK) Error("Subformula expected", fla+pptr)
		}
	}
	return Finish( sbf, dcn, ns );
}




int Match( char *start )
{
	int i;

	for ( i = 1; start[i] && start[i] != ')'; i++ )
	if ( start[i] == '(' ) i += Match(start+i)-1;
	if ( *start == '.' ) return i;
	if ( !start[i] ) Error("Unmatched left parenthesis", start)
	return i+1;
}




int Finish( int subf[], char conn[], int tot )
{
	int i, m = 0;
	char *sk;

	if ( !tot ) return *subf;
	sk = dyads;
	for ( i = 0; i < tot; i++ )
	if ( dyadic(conn[i]) >= sk ) {
		m = i;
		sk = dyadic(conn[i]);
	}
	i = m+1;
	return	Loc(	conn[m],
			Finish( subf, conn, m ),
			Finish( subf+i, conn+i, tot-i ));
}




int Loc( char mn, int lft, int rgt )
{
	register int i;

	if ( mn == '+' )
	return Loc( '>', Loc( '>', lft, 1 ), rgt );

	if ( mn == '*' )
	return Loc( '>', Loc( '>', Loc( '&', lft, rgt ), 1 ), 1 );

	for ( i = 1; wff[i].sym; i++ )
	if ( wff[i].sym == mn && wff[i].left == lft && wff[i].right == rgt )
	return i;
	wff[i].sym = mn;
	wff[i].left = lft;
	wff[i].right = rgt;
	if ( i == MAXSUBS-1) Abort("Too many subformulas");
	return i;
}
