/*				ML_TREES.C
*
*	This file contains the tree-manipulation functions for
*	the program minlog.c which is a theorem prover for
*	minimal logic.  The functions given here are to check
*	whether a given sequent is in a given database and to
*	add a similarly given sequent to a database.
*
*	Also in this file are the functions pertaining to the
*	list of ->|- applications in the current branch.
*	These might be re-done using a linked list or even a 
*	binary tree instead of an array if this proves useful.
*/


#include "minlog.h"

#define equals(S1,S2) ( S1->lowbits == S2->lowbits && S1->highbits == S2->highbits )
#define before(S1,S2) ( S1->lowbits < S2->lowbits || \
	(S1->lowbits == S2->lowbits && S1->highbits < S2->highbits))


boolean got( TREE *root, set Gamma, int conclusion )
{
	if ( !root ) return( false );
	if ( equals( root->prem, Gamma) && root->conc == conclusion )
		return( true );
	if ( root->conc < conclusion ||
		( root->conc == conclusion && before( root->prem, Gamma ) ))
		return( got( root->rightsub, Gamma, conclusion ));
	return( got( root->leftsub, Gamma, conclusion ));
}



void add( boolean good, set Gamma, int conclusion )
{
	int i;

	if ( good )
	{	if ( valids )	r_add( valids, Gamma, conclusion );
		else		valids = newguy( Gamma, conclusion );
	}
	else
	{	if ( invalids )	r_add(invalids, Gamma, conclusion );
		else		invalids = newguy( Gamma, conclusion );
	}
}






void r_add( TREE *root, set Gamma, int conclusion )
{
	if ( root->conc < conclusion ||
		( root->conc == conclusion && before( root->prem, Gamma ) ))
	{	if ( root->rightsub )
			r_add( root->rightsub, Gamma, conclusion );
		else
			root->rightsub = newguy( Gamma, conclusion );
	}
	else
	{	if ( root->leftsub )
			r_add( root->leftsub, Gamma, conclusion );
		else
			root->leftsub = newguy( Gamma, conclusion );

	}
}



TREE *newguy( set Gamma, int conclusion )
{
	TREE *ng;

	ng = (TREE*) malloc(sizeof(TREE));
	ng->prem = (set) malloc(sizeof(SET));
	copyset( Gamma, ng->prem );
	ng->conc = conclusion;
	ng->leftsub = ng->rightsub = 0;
	return( ng );
}




boolean inbranch( set Gamma, int conclusion )
{
	int i;

	for ( i = 0; i < ibmax; i++ )
	if ( equals( (p_branch+i), Gamma ) && c_branch[i] == conclusion )
		return( true );
	return( false );
}




void putinbranch( set Gamma, int conclusion )
{
	if ( ibmax == IBMAX ) Abort("Branch too long.");
	copyset( Gamma, (p_branch+ibmax) );
	c_branch[ibmax++] = conclusion;
}



void outbranch( set Gamma, int conclusion )
{
	int i, j;

	for ( i = 0; i < ibmax; i++ )
	if ( equals( (p_branch+i), Gamma ) && c_branch[i] == conclusion )
	{	for ( j = i; j < ibmax; j++ )
		{	copyset( (p_branch+j), (p_branch+j+1) );
			c_branch[j] = c_branch[j+1];
		}
		ibmax--;
		return;
	}
}